from unittest import mock

import pytest
from plotly.graph_objs import Figure

from apple_health_parser.exceptions import InvalidImageFormat
from apple_health_parser.plot.plots import Plot
from apple_health_parser.utils.parser import Parser


class TestPlot:
    def test_get_figure(self, parser: Parser) -> None:
        records = parser.get_flag_records(flag="HKQuantityTypeIdentifierHeartRate")

        plot = Plot(data=records, year=2024)
        fig = plot._get_figure()
        assert isinstance(fig, Figure)

        plot = Plot(data=records, year=2024, operation="count")
        fig = plot._get_figure()
        assert isinstance(fig, Figure)

        heatmap = Plot(
            data=records,
            year=2024,
            operation="sum",
            heatmap=True,
        )
        fig = heatmap._get_figure()
        assert isinstance(fig, Figure)

    def test_plot(self, parser: Parser) -> None:
        records = parser.get_flag_records(flag="HKQuantityTypeIdentifierHeartRate")

        plot = Plot(data=records, year=2024)
        fig = plot.plot(show=False, save=False)
        assert isinstance(fig, Figure)

        with (
            mock.patch("plotly.graph_objs.Figure.show") as mock_show,
            mock.patch("plotly.graph_objs.Figure.write_html") as mock_write_html,
            mock.patch("plotly.graph_objs.Figure.write_image") as mock_write_image,
        ):
            plot.plot(show=True, save=False)
            mock_show.assert_called_once()

            plot.plot(show=False, save=True, format="html")
            mock_write_html.assert_called_once()

            plot.plot(show=False, save=True, format="png")
            mock_write_image.assert_called_once()

    def test_plot_invalid_image_format(self, parser: Parser) -> None:
        records = parser.get_flag_records(flag="HKQuantityTypeIdentifierHeartRate")

        overview = Plot(data=records, year=2024)
        fmt = "tiff"
        with pytest.raises(InvalidImageFormat):
            overview.plot(show=False, save=True, format=fmt)



from dataclasses import dataclass
from importlib import resources

import yaml

from apple_health_parser.config.definitions import Operations, PlotType


@dataclass(frozen=False)
class MetricDefinition:
    """
    Dataclass to hold the metadata for each health metric, including:
        - `axis_label`: The label for the y-axis in plots
        - `color`: The color to use for the metric in plots
        - `description`: A brief description of the metric
        - `flag`: The unique identifier for the metric in the Apple Health data
        - `marker_color`: The color to use for markers in plots
        - `name`: A human-readable name for the metric
        - `operation`: The type of aggregation operation to perform (e.g. sum, mean)
        - `plot_type`: The type of plot to use for the metric (e.g. bar, line)
        - `summary_template`: The template for summarizing the metric
        - `unit`: The unit of measurement for the metric (e.g. "km", "min")
    """

    axis_label: str
    color: str
    description: str
    flag: str
    marker_color: str
    name: str
    operation: Operations
    plot_type: PlotType
    summary_template: str
    unit: str


metrics_file = (
    resources.files("apple_health_parser.scripts.recap.metrics") / "metrics.yaml"
)
metrics_content = yaml.safe_load(metrics_file.read_text())
METRIC_DEFINITIONS: dict[str, MetricDefinition] = {
    flag: MetricDefinition(**metadata) for flag, metadata in metrics_content.items()
}



import io

import pandas as pd
import plotly.graph_objects as go

from apple_health_parser.config.definitions import PlotType


def get_recap_plots(
    flag: str,
    df_merged: pd.DataFrame,
    df_day: pd.DataFrame,
    week_labels: list[str],
    month_annotations: list[dict],
    axis_label: str,
    plot_type: PlotType,
    metric_color: str,
) -> tuple[io.BytesIO, io.BytesIO]:
    """
    Generate the bar chart and box plot for the year recap.

    Args:
        flag (str): Health data type identifier to determine the appropriate plot settings (e.g. "HKQuantityTypeIdentifierDistanceWalkingRunning")
        df_merged (pd.DataFrame): DataFrame containing the weekly data
        df_day (pd.DataFrame): DataFrame containing the daily data for the box plot
        week_labels (list[str]): List of week labels for the x-axis
        month_annotations (list[dict]): List of annotation dicts for month labels
        axis_label (str): Label for the y-axis (e.g. "Distance (km)")
        plot_type (PlotType): Type of plot to generate (e.g. PlotType.BAR)
        metric_color (str): Hex color code for the metric (e.g. "#fe5841")

        Returns:
            tuple[io.BytesIO, io.BytesIO]: Tuple containing the SVG image data for the bar chart and box plot
                - Bar chart image data
                - Box plot image data
    """

    fig = go.Figure()

    if plot_type == PlotType.BAR:
        bar_kwargs = dict(
            x=week_labels,
            y=df_merged["value"],
            name=axis_label,
        )

        # Time in daylight gets a special color mapping based on the value
        if flag == "HKQuantityTypeIdentifierTimeInDaylight":
            bar_kwargs["marker"] = dict(
                color=df_merged["value"],
                colorscale=[[0, "#1a237e"], [1, "#ffd700"]],
            )
        else:
            # Other metrics use a single color
            bar_kwargs["marker_color"] = metric_color

        fig.add_trace(go.Bar(**bar_kwargs))

    elif plot_type == PlotType.LINE:
        fig.add_trace(
            go.Scatter(
                x=week_labels,
                y=df_merged["value"],
                mode="lines",
                name=axis_label,
                line_color=f"{metric_color}",
                line_shape="spline",
                marker=dict(size=6),
            )
        )

    fig.update_layout(
        width=900,
        height=450,
        xaxis=dict(
            tickangle=-90,
            tickvals=week_labels,
            ticktext=week_labels,
        ),
        yaxis_title=axis_label,
        template="simple_white",
        showlegend=False,
        margin=dict(l=0, r=0, t=0, b=60),  # Add bottom margin for annotations
        annotations=month_annotations,
    )

    image_data = fig.to_image(format="svg")
    image = io.BytesIO(image_data)

    fig_box = go.Figure()
    fig_box.add_trace(
        go.Box(
            x=df_day,
            marker_color=f"{metric_color}",
            boxpoints="all",
            pointpos=-2,
            notched=True,
            marker=dict(size=3),
        )
    )

    fig_box.update_layout(
        xaxis=dict(title=axis_label),
        yaxis=dict(visible=False),
        width=600,
        height=100,
        template="simple_white",
        showlegend=False,
        margin=dict(l=0, r=0, t=0, b=20),
    )
    box_image_data = fig_box.to_image(format="svg")
    box_image = io.BytesIO(box_image_data)

    return image, box_image



from datetime import datetime

from plotly.graph_objects import Figure, Scatter

from apple_health_parser.config.definitions import SleepColors
from apple_health_parser.interfaces.plot_interface import PlotInterface
from apple_health_parser.models.parsed import ParsedData
from apple_health_parser.models.records import SleepType


class SleepPlot(PlotInterface):
    """
    Plot the parsed sleep data.
    """

    def __init__(
        self, data: ParsedData, year: int, timerange: tuple[str, str] | None = None
    ):
        """
        Initialize the SleepPlot.

        Args:
            data (ParsedData): Parsed data object containing sleep records.
            year (int): Year for which the data is plotted.
            timerange (tuple, optional): Start and end date for the plot in ISO format.
                If provided, the data will be filtered to include only records within this range.
                Must be a tuple of two strings in ISO format (e.g. `("2024-03-01T20:00:00+00:00", "2024-03-02T08:00:00+00:00")`).
                Defaults to None, which means no filtering is applied.
        """
        super().__init__(data=data, year=year)

        if timerange is not None:
            # Validate timerange
            if not isinstance(timerange, tuple) or len(timerange) != 2:
                raise ValueError("timerange must be a tuple of two date strings.")
            if not all(isinstance(date, str) for date in timerange):
                raise ValueError(
                    "Both elements of timerange must be strings in ISO format."
                )

            # Convert strings to datetime objects
            start_dt, end_dt = (datetime.fromisoformat(date) for date in timerange)
            timerange_iso: tuple[datetime, datetime] = (start_dt, end_dt)

            # Filter the dataframe based on the timerange
            self.dataframe = self.dataframe[
                (self.dataframe["start_date"] >= timerange_iso[0])
                & (self.dataframe["end_date"] <= timerange_iso[1])
            ]

    def _get_figure(self) -> Figure:
        """
        Get the plotly figure for sleep data.

        Returns:
            Figure: Figure object
        """
        colors = {
            SleepType.IN_BED: SleepColors.in_bed,
            SleepType.AWAKE: SleepColors.awake,
            SleepType.CORE: SleepColors.core,
            SleepType.DEEP: SleepColors.deep,
            SleepType.REM: SleepColors.rem,
            SleepType.UNSPECIFIED: SleepColors.unset,
        }

        fig = Figure()
        for row in self.dataframe.itertuples():
            fig.add_trace(
                Scatter(
                    x=[row.start_date, row.end_date],
                    y=[row.value, row.value],
                    name=row.value,
                    showlegend=False,
                    mode="lines",
                    line_shape="hvh",
                    line=dict(
                        color=colors.get(SleepType(row.value), "black"), width=10
                    ),
                    hoverinfo="text",
                    hovertext=f"{row.start_date} - {row.end_date}<br>Sleep stage: {row.value}",
                )
            )

        fig.update_layout(
            xaxis_title="Date",
            yaxis_title=self.psets.title_yaxis,
            legend_title_text=self.psets.legend,
            template="simple_white",
        )

        return fig



import pandas as pd
import pytest

from apple_health_parser.exceptions import (
    InvalidOperation,
    InvalidSource,
    MissingYear,
)
from apple_health_parser.utils.parser import Parser
from apple_health_parser.utils.preprocessor import Preprocessor


class TestPreprocessor:
    def test_exceptions(self, parser: Parser) -> None:
        data = parser.get_flag_records("HKQuantityTypeIdentifierActiveEnergyBurned")
        year = 2024
        source = "Alexandre's Apple Watch"
        operation = "sum"
        heatmap = False

        with pytest.raises(MissingYear):
            Preprocessor(
                data, year=2020, source=source, operation=operation, heatmap=heatmap
            )

        with pytest.raises(InvalidOperation):
            Preprocessor(
                data, year=year, source=source, operation="test", heatmap=heatmap
            )

        with pytest.raises(InvalidSource):
            Preprocessor(
                data,
                year=year,
                source="Invalid Source",
                operation=operation,
                heatmap=heatmap,
            )

    def test_get_dataframe(self, parser: Parser) -> None:
        data = parser.get_flag_records("HKQuantityTypeIdentifierActiveEnergyBurned")
        year = 2024
        source = "Alexandre's Apple Watch"
        operation = "sum"
        heatmap = False
        preprocessor = Preprocessor(data, year, source, operation, heatmap)
        df = preprocessor.get_dataframe()

        assert isinstance(df, pd.DataFrame)
        assert df.shape == (2, 2)

        heatmap = True
        preprocessor = Preprocessor(data, year, source, operation, heatmap)
        df = preprocessor.get_dataframe()
        assert df.shape == (1, 2)

    def test_get_heatmap(self, parser: Parser) -> None:
        data = parser.get_flag_records("HKQuantityTypeIdentifierActiveEnergyBurned")
        year = 2024
        source = "Alexandre's Apple Watch"
        operation = "sum"
        heatmap = False
        preprocessor = Preprocessor(data, year, source, operation, heatmap)
        df = preprocessor.get_dataframe()
        heatmap = Preprocessor.get_heatmap(df)

        assert isinstance(heatmap, pd.DataFrame)
        assert heatmap.shape == (1, 2)
        assert heatmap.columns.name == "day"
        assert heatmap.index.name == "month"



from unittest import mock

import pytest
from plotly.graph_objs import Figure

from apple_health_parser.exceptions import (
    InvalidImageFormat,
    InvalidOverviewType,
    MissingFlag,
)
from apple_health_parser.plot.overviews import Overview
from apple_health_parser.utils.parser import Parser


@pytest.fixture
def flags() -> list[str]:
    return [
        "HKQuantityTypeIdentifierActiveEnergyBurned",
        "HKQuantityTypeIdentifierAppleExerciseTime",
        "HKQuantityTypeIdentifierAppleStandTime",
    ]


class TestOverviews:
    def test_validate(self, parser: Parser, flags: list[str]) -> None:
        records = parser.get_flag_records(flag=flags)

        with pytest.raises(InvalidOverviewType):
            Overview(data=records, overview_type="fake-overview", year=2024)

        with pytest.raises(MissingFlag):
            Overview(data=records, overview_type="body", year=2024)

    def test_get_figure(self, parser: Parser, flags: list[str]) -> None:
        records = parser.get_flag_records(flag=flags)

        overview = Overview(data=records, overview_type="activity", year=2024)
        fig = overview._get_figure()
        assert isinstance(fig, Figure)

    def test_plot(self, parser: Parser, flags: list[str]) -> None:
        records = parser.get_flag_records(flag=flags)

        overview = Overview(data=records, overview_type="activity", year=2024)
        fig = overview.plot(show=False, save=False)
        assert isinstance(fig, Figure)

        with (
            mock.patch("plotly.graph_objs.Figure.show") as mock_show,
            mock.patch("plotly.graph_objs.Figure.write_html") as mock_write_html,
            mock.patch("plotly.graph_objs.Figure.write_image") as mock_write_image,
        ):
            overview.plot(show=True, save=False)
            mock_show.assert_called_once()

            overview.plot(show=False, save=True, format="html")
            mock_write_html.assert_called_once()

            overview.plot(show=False, save=True, format="png")
            mock_write_image.assert_called_once()

    def test_plot_invalid_image_format(self, parser: Parser, flags: list[str]) -> None:
        records = parser.get_flag_records(flag=flags)

        overview = Overview(data=records, overview_type="activity", year=2024)
        fmt = "tiff"
        with pytest.raises(InvalidImageFormat):
            overview.plot(show=False, save=True, format=fmt)



import typing

import pandas as pd

from apple_health_parser.config.definitions import Operations
from apple_health_parser.scripts.recap.metrics.definitions import (
    METRIC_DEFINITIONS,
)


@typing.no_type_check
def get_recap_stats(
    df: pd.DataFrame,
    flag: str,
) -> tuple[pd.DataFrame, str, tuple[float, float, float, float, float]]:
    """
    Compute daily distance statistics and return a daily summary DataFrame, a formatted statistics text, and quantiles.

    Args:
        df (pd.DataFrame): DataFrame containing the original distance records with 'start_date' and 'value' columns
        flag (str): Health data type identifier to determine the appropriate summary template (e.g. "HKQuantityTypeIdentifierDistanceWalkingRunning")

    Returns:
        tuple:
            - df_day (pd.DataFrame): DataFrame with daily data
            - stats_text (str): Formatted string summarizing the distance statistics
            - quantiles (tuple): Tuple containing the min, Q1, median, Q3, and max of daily distances
    """
    df["day"] = df["start_date"].dt.to_period("D")

    operation = METRIC_DEFINITIONS[flag].operation
    if operation == Operations.SUM:
        df_day = df.groupby("day")["value"].sum()
    elif operation == Operations.MEAN:
        df_day = df.groupby("day")["value"].mean()

    dist_count = df_day.count()
    dist_avg = df_day.mean()
    dist_total = df_day.sum()
    q_min = df_day.min()
    q_max = df_day.max()
    q1 = df_day.quantile(0.25)
    q2 = df_day.median()
    q3 = df_day.quantile(0.75)

    stats_text = METRIC_DEFINITIONS[flag].summary_template.format(
        dist_total=dist_total,
        dist_avg=dist_avg,
        dist_median=q2,
        dist_min=q_min,
        dist_max=q_max,
        dist_count=dist_count,
    )

    return df_day, stats_text, (q_min, q1, q2, q3, q_max)



import logging
import os
import sys
import warnings
from datetime import date
from pathlib import Path

import click

from apple_health_parser.scripts.recap.metrics.definitions import (
    METRIC_DEFINITIONS,
)
from apple_health_parser.scripts.recap.recap_data import get_recap_data
from apple_health_parser.scripts.recap.recap_pdf import PdfSectionData, get_recap_report
from apple_health_parser.scripts.recap.recap_plot import get_recap_plots
from apple_health_parser.scripts.recap.recap_stats import get_recap_stats
from apple_health_parser.utils.logging import logger
from apple_health_parser.utils.parser import Parser

# Set logging level for fpdf to ERROR to suppress warnings
logging.getLogger("fpdf").setLevel(logging.ERROR)

# Ignore warnings about timezone information dropping when converting to PeriodArray/Index representation
warnings.filterwarnings(
    "ignore",
    category=UserWarning,
    message="Converting to PeriodArray/Index representation will drop timezone information.",
)


def open_pdf(filepath: Path) -> None:
    """
    Open the generated PDF report using the default PDF viewer based on the operating system.

    Args:
        filepath (Path): The path to the PDF file to open

    Raises:
        RuntimeError: If the platform is unsupported
    """
    if sys.platform.startswith("darwin"):
        os.system(f'open "{filepath}"')
    elif sys.platform.startswith("win"):
        os.startfile(filepath)
    elif sys.platform.startswith("linux"):
        os.system(f'xdg-open "{filepath}"')
    else:
        raise RuntimeError(f"Unsupported platform: {sys.platform}")


@click.command(no_args_is_help=True)
@click.option("--zip_file", help="Path to the Apple Health export.zip file")
@click.option(
    "--year",
    default=date.today().year,
    help="Year to filter the data for, defaults to current year",
)
def main(zip_file: str, year: int = date.today().year) -> None:
    """
    CLI to export the year recap from the Apple Health export file to a PDF file.

    Args:
        zip_file (str): Path to the Apple Health export.zip file
        year (int, optional): Year to filter the data for, defaults to current year
    """
    logger.info(
        click.style(f"Generating year recap report - {year}...", fg="green", bold=True)
    )

    # Parse data
    parser = Parser(export_file=zip_file, overwrite=True, verbose=True)

    # Get data and prepare for plotting
    pdf_section_data: dict[str, PdfSectionData] = {}
    for n, (flag, metadata) in enumerate(METRIC_DEFINITIONS.items()):
        try:
            logger.info(
                click.style(
                    f"Processing metric {n + 1}/{len(METRIC_DEFINITIONS)}: {flag}",
                    fg="blue",
                )
            )

            if flag not in parser.flags:
                logger.warning(
                    click.style(
                        f"\nFlag not found in data, skipping: {flag}", fg="yellow"
                    )
                )
                continue

            df, df_merged, week_labels, month_annotations = get_recap_data(
                parser, year, flag, metadata.operation
            )

            df_day, stats_text, (q_min, q1, q2, q3, q_max) = get_recap_stats(df, flag)

            image, box_image = get_recap_plots(
                flag,
                df_merged,
                df_day,
                week_labels,
                month_annotations,
                metadata.axis_label,
                metadata.plot_type,
                metadata.color,
            )

            pdf_section_data[flag] = PdfSectionData(
                year=year,
                image_data=image,
                box_image_data=box_image,
                stats_text=stats_text,
                quantiles=(q_min, q1, q2, q3, q_max),
            )

        except Exception as e:
            logger.warning(
                click.style(f"Error processing metric {flag}: {e}", fg="yellow")
            )

    try:
        get_recap_report(pdf_section_data)
        pdf_filepath = Path("report.pdf").resolve()
        logger.info(click.style(f"Report generated: {pdf_filepath}", underline=True))
        open_pdf(pdf_filepath)

    except StopIteration:
        logger.error(
            click.style(
                "No data found to generate the report. Please check your data and/or CLI command, and try again.",
                fg="red",
            )
        )


if __name__ == "__main__":
    main()



from pathlib import Path
from tempfile import TemporaryDirectory
from unittest import mock
from zipfile import ZipFile

import lxml.etree as ET

from apple_health_parser.utils.loader import Loader


def test_extract_zip() -> None:
    with mock.patch("click.confirm", return_value=True):
        with TemporaryDirectory() as temp_dir:
            # Trigger deletion of previous export warning
            for _ in range(2):
                zip_path = Path(temp_dir) / "test.zip"

                with ZipFile(zip_path, "w") as zip_file:
                    zip_file.writestr("apple_health_export/export.xml", "test data")

                result = Loader.extract_zip(str(zip_path), temp_dir)
                expected = Path(temp_dir) / "apple_health_export/export.xml"

                assert result.resolve() == expected.resolve()
                assert result.exists()


@mock.patch("apple_health_parser.utils.loader.Loader._log_metadata")
def test_read_xml(mock_log_metadata, xml_file: Path) -> None:
    loader = Loader()
    result = loader.read_xml(xml_file=xml_file)
    mock_log_metadata.assert_called_once()

    assert isinstance(result, list)
    assert all(isinstance(item, ET.Element) for item in result)


@mock.patch("apple_health_parser.utils.logging.logger.info")
def test_log_metadata(mock_logger_info, xml_file: Path, root: ET.Element) -> None:
    with open(xml_file, "r") as _:
        Loader._log_metadata(root)
        mock_logger_info.assert_called()
        assert mock_logger_info.call_count == 7



from datetime import date
from pathlib import Path
from unittest import mock

import lxml.etree as ET
import pandas as pd
import pytest

from apple_health_parser.exceptions import (
    InvalidFileFormat,
    InvalidFlag,
    MissingRecords,
)
from apple_health_parser.models.parsed import ParsedData
from apple_health_parser.models.records import HealthData, HeartRateData
from apple_health_parser.utils.parser import Parser


class TestParser:
    def test_init(self, xml_file: str, export_file: str, tmp_path: Path) -> None:
        with (
            mock.patch(
                "apple_health_parser.utils.loader.Loader.extract_zip",
                return_value=xml_file,
            ) as mock_extract_zip,
            mock.patch(
                "apple_health_parser.utils.parser.Parser._get_records",
                return_value=[],
            ) as mock_get_records,
        ):
            parser = Parser(export_file=export_file, output_dir=tmp_path)

            assert parser.xml_file == mock_extract_zip.return_value
            assert parser.records == mock_get_records.return_value

            mock_extract_zip.assert_called_once()
            mock_get_records.assert_called_once()

    def test_get_records(self, parser: Parser) -> None:
        expected_flags = [
            "HKQuantityTypeIdentifierHeartRate",
            "HKQuantityTypeIdentifierActiveEnergyBurned",
        ]
        records = parser._get_records()

        assert isinstance(records, dict)
        assert all(flag in records for flag in expected_flags)
        assert all(
            isinstance(rec, ET.Element)
            for rec_list in records.values()
            for rec in rec_list
        )

    def test_build_models(self, parser: Parser) -> None:
        heart_rate_models = parser._build_models("HKQuantityTypeIdentifierHeartRate")
        active_energy_models = parser._build_models(
            "HKQuantityTypeIdentifierActiveEnergyBurned"
        )
        with pytest.raises(InvalidFlag):
            parser._build_models("HKQuantityTypeIdentifierStepCount")

        assert len(heart_rate_models) == 2
        assert isinstance(heart_rate_models[0], HeartRateData)
        assert isinstance(heart_rate_models[1], HeartRateData)
        assert heart_rate_models[0].type == "HKQuantityTypeIdentifierHeartRate"
        assert heart_rate_models[1].type == "HKQuantityTypeIdentifierHeartRate"
        assert heart_rate_models[0].value == 74
        assert heart_rate_models[1].value == 75
        assert heart_rate_models[0].unit == "count/min"
        assert heart_rate_models[1].unit == "count/min"
        assert heart_rate_models[0].motion_context == "Unset"
        assert heart_rate_models[1].motion_context == None
        assert heart_rate_models[0].source_version == "10.2"
        assert heart_rate_models[1].source_version == "10.2"
        assert len(active_energy_models) == 2
        assert isinstance(active_energy_models[0], HealthData)
        assert (
            active_energy_models[0].type == "HKQuantityTypeIdentifierActiveEnergyBurned"
        )
        assert active_energy_models[0].value == 0.12
        assert active_energy_models[1].value == 0.08
        assert active_energy_models[0].unit == "kcal"
        assert active_energy_models[0].source_version == "10.2"

    def test_get_dates(self, parser: Parser) -> None:
        active_energy_models = parser._build_models(
            "HKQuantityTypeIdentifierActiveEnergyBurned"
        )
        dates = parser._get_dates(active_energy_models)
        actual_dates = {method() for method in dates}

        assert actual_dates == {date(2024, 1, 1), date(2024, 1, 2)}

    def test_map_record_keys_to_flags(self, parser: Parser) -> None:
        flags = [
            "HKCategoryTypeIdentifierSleepAnalysis",
            "HKQuantityTypeIdentifierActiveEnergyBurned",
            "HKQuantityTypeIdentifierAppleExerciseTime",
            "HKQuantityTypeIdentifierAppleStandTime",
            "HKQuantityTypeIdentifierHeartRate",
        ]
        record_keys = {
            "device",
            "type",
            "startDate",
            "value",
            "sourceName",
            "endDate",
            "creationDate",
            "sourceVersion",
            "unit",
        }
        flag_map = parser._map_record_keys_to_flags()
        # Check if the flags are in the flag_map
        assert sorted(flag_map.keys()) == flags
        # Check if the record keys are as expected
        assert all(record_keys == i for i in flag_map.values())

    def test_get_flag_records(self, parser: Parser) -> None:
        with (
            mock.patch(
                "apple_health_parser.utils.parser.Parser._build_models"
            ) as mock_build_models,
            mock.patch(
                "apple_health_parser.utils.parser.Parser._get_dates"
            ) as mock_get_dates,
            mock.patch(
                "apple_health_parser.utils.parser.Parser.get_sources"
            ) as mock_get_sources,
        ):
            result = parser.get_flag_records("HKQuantityTypeIdentifierHeartRate")

            mock_get_dates.assert_called_once()
            mock_build_models.assert_called_once()
            mock_get_sources.assert_called_once()
            assert isinstance(result, ParsedData)

            result = parser.get_flag_records(
                [
                    "HKQuantityTypeIdentifierHeartRate",
                    "HKQuantityTypeIdentifierActiveEnergyBurned",
                ]
            )

            assert mock_get_dates.call_count == 3
            assert mock_build_models.call_count == 3
            assert mock_get_sources.call_count == 3
            assert isinstance(result, dict)

    def test_write_csv(self, parser: Parser) -> None:
        with mock.patch.object(pd.DataFrame, "to_csv") as mock_to_csv:
            data = parser.get_flag_records("HKQuantityTypeIdentifierHeartRate")
            parser.write_csv(data=data, filename="test.csv")
            mock_to_csv.assert_called_once()

            with pytest.raises(InvalidFileFormat):
                parser.write_csv(data=data, filename="test.pdf")
                mock_to_csv.assert_not_called()

            data.records = None
            with pytest.raises(MissingRecords):
                parser.write_csv(data=data, filename="test.csv")
                mock_to_csv.assert_not_called()

    def test_export(self, parser: Parser, tmp_path: Path) -> None:
        with (
            mock.patch(
                "apple_health_parser.utils.parser.Parser.get_flag_records"
            ) as mock_get_flag_records,
            mock.patch(
                "apple_health_parser.utils.parser.Parser.write_csv"
            ) as mock_write_csv,
        ):
            parser.export(dir_name=tmp_path)

            assert mock_get_flag_records.call_count == 5
            assert mock_write_csv.call_count == 5

    def test_parsed(self, parser: Parser) -> None:
        flag = "HKQuantityTypeIdentifierHeartRate"
        records = parser.get_flag_records(flag=flag)
        str_records = str(records)

        assert "ParsedData" in str_records
        assert flag in str_records
        assert all(
            item in str_records for item in ["Flag", "Sources", "Dates", "Records"]
        )



import subprocess


def test_cli_help():
    """Very simple test - just check that CLI help works without errors."""
    result = subprocess.run(
        ["python", "-m", "apple_health_parser.scripts.export.main", "--help"],
        capture_output=True,
        text=True,
    )
    assert result.returncode == 0
    assert "CLI to export data" in result.stdout


def test_cli_without_args():
    """Test CLI shows help when no arguments provided."""
    result = subprocess.run(
        ["python", "-m", "apple_health_parser.scripts.export.main"],
        capture_output=True,
        text=True,
    )
    assert result.returncode == 2
    assert "Usage:" in result.stderr


def test_cli_invalid_option():
    """Test CLI handles invalid options gracefully."""
    result = subprocess.run(
        ["python", "-m", "apple_health_parser.scripts.export.main", "--invalid-option"],
        capture_output=True,
        text=True,
    )
    assert result.returncode != 0



import typing

import pandas as pd

from apple_health_parser.config.definitions import Operations
from apple_health_parser.exceptions import MissingYear
from apple_health_parser.models.parsed import ParsedData
from apple_health_parser.utils.parser import Parser


@typing.no_type_check
def get_recap_data(
    parser: Parser, year: int, flag: str, operation: Operations
) -> tuple[pd.DataFrame, pd.DataFrame, list[str], list[dict]]:
    """
    Parse Apple Health data and prepare weekly summary, labels, and month annotations for plotting.

    Args:
        parser (Parser): Instance of the Parser class
        year (int): Year to filter the data for
        flag (str): Health data type identifier to extract (e.g. "HKQuantityTypeIdentifierDistanceWalkingRunning")
        operation (Operations): Operation type for the data (e.g. "sum")

    Returns:
        tuple:
            - df (pd.DataFrame): Original filtered DataFrame for further statistics
            - df_merged (pd.DataFrame): Weekly summary DataFrame with all weeks for the year
            - week_labels (pd.Series): Series of week labels for plotting
            - month_annotations (list[dict]): List of annotation dicts for month labels
    """
    # Get records for the specified flag
    parsed: ParsedData = parser.get_flag_records(flag)
    df = parsed.records

    years = df.start_date.dt.year.unique().tolist()
    if year not in years:
        raise MissingYear(year, years)

    # Ensure start_date is datetime
    df["start_date"] = pd.to_datetime(df["start_date"])

    # Filter by year
    df = df[df["start_date"].dt.year == year]

    # Compute week start
    df["week_start"] = df["start_date"].dt.to_period("W").dt.start_time

    # Compute week start and aggregate
    if operation == Operations.SUM:
        df_week = df.groupby("week_start", as_index=False)["value"].sum()
    elif operation == Operations.MEAN:
        df_week = df.groupby("week_start", as_index=False)["value"].mean()
    else:
        raise ValueError(f"Unsupported operation: {operation}")

    # Generate all weeks for the year
    weeks = pd.date_range(start=f"{year}-01-01", end=f"{year}-12-31", freq="W-MON")

    # Merge with all weeks to ensure we have a complete timeline, filling missing weeks with NaN
    df_merged = pd.DataFrame({"week_start": weeks}).merge(
        df_week,
        on="week_start",
        how="left",
    )
    df_merged["value"] = df_merged["value"].fillna(pd.NA)

    # Drop weeks that are outside the specified year (e.g., last week of previous year or first week of next year)
    isocal = df_merged["week_start"].dt.isocalendar()
    df_merged["week"] = isocal.week
    df_merged["month"] = df_merged["week_start"].dt.strftime("%b")
    df_merged = df_merged[isocal.year == year]

    # Labels for plotting
    week_labels = df_merged["week"].astype(str).str.zfill(2)

    # Find first week index for each month for annotation
    month_week_index = list(
        df_merged.drop_duplicates("month")[["month", "week"]]
        .sort_values("week")
        .itertuples(index=False, name=None)
    )
    month_annotations = [
        dict(
            x=week_num,
            y=-0.12,  # y-offset below the x-axis
            xref="x",
            yref="paper",
            text=month,
            showarrow=False,
            font=dict(size=14),
            align="left",
        )
        for month, week_num in month_week_index
    ]

    return df, df_merged, week_labels, month_annotations



import click

from apple_health_parser.utils.logging import logger
from apple_health_parser.utils.parser import Parser


@click.command(no_args_is_help=True)
@click.option("--zip_file", help="Path to the Apple Health export.zip file")
@click.option(
    "--dir_name",
    default=".",
    help="Directory to export the CSV files to, defaults to current directory",
)
def main(zip_file: str, dir_name: str) -> None:
    """
    CLI to export data from the Apple Health export file to CSV files.
    """
    logger.info(click.style("Apple Health Parser", bg="blue", fg="white", bold=True))

    parser = Parser(export_file=zip_file, overwrite=True, verbose=True)

    # Export all data
    parser.export(dir_name=dir_name)


if __name__ == "__main__":
    main()



from dataclasses import dataclass
from importlib.metadata import version
from io import BytesIO

from fpdf import FPDF

from apple_health_parser.scripts.recap.metrics.definitions import (
    METRIC_DEFINITIONS,
)


@dataclass
class PdfSectionData:
    year: int
    image_data: BytesIO
    box_image_data: BytesIO
    stats_text: str
    quantiles: tuple[float, float, float, float, float]


class PDF(FPDF):
    def footer(self):
        self.set_y(-15)
        self.set_font("Arial", style="I", size=7)
        self.cell(0, 10, f"Page {self.page_no()}/{{nb}}", align="C")
        self.set_font("Arial", style="", size=7)
        self.set_text_color(100)
        footer_text = (
            f"Generated with **Apple Health Parser** {version('apple-health-parser')}"
        )
        self.cell(0, 10, footer_text, align="R", markdown=True)


def get_recap_report(
    pdf_section_data: dict[str, PdfSectionData],
) -> None:
    """
    Generate a PDF report for the year recap.

    Args:
        pdf_section_data (dict[str, PdfSectionData]): Data for each section of the report, keyed by flag
    """
    pdf = PDF(orientation="P", unit="mm", format="A4")
    pdf.author = "Apple Health Parser"
    pdf.title = f"Apple Health Year Recap - {pdf_section_data[next(iter(pdf_section_data))].year}"

    # Text block width
    pdf.set_left_margin(20)
    pdf.set_right_margin(20)
    WIDTH = (pdf.w - pdf.l_margin - pdf.r_margin) * 0.8
    TABLE_WIDTH = WIDTH * 0.8

    for page, (flag, section_data) in enumerate(pdf_section_data.items()):
        metric_metadata = METRIC_DEFINITIONS[flag]

        year, image_data, box_image_data, stats_text, quantiles = (
            section_data.year,
            section_data.image_data,
            section_data.box_image_data,
            section_data.stats_text,
            section_data.quantiles,
        )

        pdf.add_page()

        if page == 0:
            pdf.set_font("Arial", "B", 36)
            pdf.ln(10)
            pdf.cell(10, 10, "Year Recap")
            pdf.ln(14)
            pdf.set_font("Arial", "B", 24)
            pdf.cell(10, 10, f"{year}")
            pdf.ln(16)

        pdf.set_font("Arial", "B", 18)
        pdf.cell(0, 10, f"{metric_metadata.name}")
        pdf.ln(10)

        metric_desc = metric_metadata.description
        pdf.set_font("Arial", "I", 9)
        pdf.set_text_color(120)
        pdf.multi_cell(WIDTH, 5, metric_desc, border="T", padding=2)
        pdf.ln(12)

        pdf.image(image_data, w=WIDTH, keep_aspect_ratio=True)
        pdf.ln(10)

        pdf.set_font("Arial", size=9)
        pdf.set_text_color(1)
        pdf.multi_cell(WIDTH, 5, stats_text, markdown=True)
        pdf.ln(6)

        pdf.image(box_image_data, w=WIDTH, keep_aspect_ratio=True)
        pdf.ln(10)

        q_min, q1, q2, q3, q_max = quantiles
        table_data = (
            ("Metric", metric_metadata.unit),
            ("Min", f"{q_min:.2f} {metric_metadata.unit}"),
            ("Q1", f"{q1:.2f} {metric_metadata.unit}"),
            ("Median", f"{q2:.2f} {metric_metadata.unit}"),
            ("Q3", f"{q3:.2f} {metric_metadata.unit}"),
            ("Max", f"{q_max:.2f} {metric_metadata.unit}"),
        )

        pdf.set_font("Arial", size=7)
        pdf.set_text_color(60)
        pdf.set_draw_color(60)
        pdf.multi_cell(
            TABLE_WIDTH,
            5,
            text="__Table__ - Five-number summary",
            border="B",
            align="C",
            padding=2,
            markdown=True,
        )
        pdf.ln(4)
        pdf.set_font("Arial", size=9)
        pdf.set_text_color(1)
        pdf.set_draw_color(1)
        with pdf.table(
            align="LEFT", width=TABLE_WIDTH, borders_layout="SINGLE_TOP_LINE"
        ) as table:
            for data_row in table_data:
                row = table.row()
                for datum in data_row:
                    row.cell(datum)

        # Reset font color
        pdf.set_text_color(1)

    pdf.output("report.pdf")



from collections import defaultdict
from datetime import date
from pathlib import Path
from typing import overload

import click
import lxml.etree as ET
import pandas as pd
from pydantic import ValidationError

from apple_health_parser.decorators import timeit
from apple_health_parser.exceptions import (
    InvalidFileFormat,
    InvalidFlag,
    MissingRecords,
)
from apple_health_parser.models.parsed import ParsedData
from apple_health_parser.models.records import HealthData, HeartRateData, SleepData
from apple_health_parser.utils.loader import Loader
from apple_health_parser.utils.logging import logger

# https://pandas.pydata.org/pandas-docs/stable/user_guide/indexing.html#returning-a-view-versus-a-copy
pd.options.mode.copy_on_write = True


class Parser(Loader):
    """
    Parser class to parse the Apple Health export file.

    Subclass of the Loader class which extracts the export.zip file and logs metadata.
    """

    def __init__(
        self,
        export_file: str | Path,
        output_dir: str | Path = "data",
        verbose: bool = False,
        overwrite: bool | None = None,
    ) -> None:
        """
        Initialize the Parser class with the path to the export.zip file.

        Args:
            export_file (str): Path to the export.zip file
            output_dir (str): Directory to export the parsed data to, defaults to "data"
            verbose (bool): Flag to enable verbose logging, defaults to False
            overwrite (bool, optional): Flag to overwrite the existing data, defaults to None
        """
        if verbose is False:
            logger.propagate = False

        self.xml_file = self.extract_zip(
            zip_file=export_file, output_dir=output_dir, overwrite=overwrite
        )
        self.records = self._get_records()
        self.flags = list(self.records.keys()) if self.records else []

    @timeit
    def _get_records(self) -> defaultdict[str, list[ET.Element]]:
        """
        Get records from the Apple Health export file.
        The records are grouped by flags as keys and list of records as values.

        Returns:
            defaultdict[str, list[ET.Element]]: Records from the export.xml file
        """
        data = self.read_xml(self.xml_file)
        records: defaultdict[str, list[ET.Element]] = defaultdict(list)
        for rec in data:
            # Match record "type" to flag
            records[rec.attrib["type"]].append(rec)
        record_count = sum(len(rec) for rec in records.values())
        logger.info(
            f"Processed {len(records.keys())} flags with {record_count:,} records"
        )
        return records

    def _build_models(self, flag: str) -> list:
        """
        Build models from the records based on the flag.

        Args:
            flag (str): Flag to parse the records

        Returns:
            list: List of models based on the flag
        """
        logger.info(f"Parsing records with flag: {click.style(flag, fg='magenta')}")

        if flag not in self.flags:
            raise InvalidFlag(flag, self.flags)

        models: list[HealthData | HeartRateData] = []
        failed: dict[str, int] = {}

        for rec in self.records[flag]:
            try:
                # Heart rate records can have additional metadata (motionContext)
                if flag == "HKQuantityTypeIdentifierHeartRate":
                    if rec.find("MetadataEntry") is not None:
                        motion_context = rec.find("MetadataEntry").attrib["value"]
                    else:
                        motion_context = None
                    models.append(
                        HeartRateData(**rec.attrib, motionContext=motion_context)
                    )
                elif flag == "HKCategoryTypeIdentifierSleepAnalysis":
                    # Sleep records have additional metadata (value)
                    models.append(
                        SleepData(
                            **{
                                **rec.attrib,
                                **{
                                    "timezone": rec.find("MetadataEntry").attrib[
                                        "value"
                                    ]
                                },
                            }
                        )
                    )
                else:
                    models.append(HealthData(**rec.attrib))

            except ValidationError as exc:
                error_type = exc.errors()[0]["type"]
                loc = exc.errors()[0]["loc"]
                try:
                    failed[f"{error_type}_{loc}"] += 1
                except KeyError:
                    failed[f"{error_type}_{loc}"] = 1
                continue

        if failed:
            logger.warning(
                click.style(f"Failed to parse {len(failed)} records", bold=True)
            )

        return models

    def _get_dates(self, models: list) -> set[date]:
        """
        Get unique month and year combinations from the models.

        Args:
            models (list): List of models

        Returns:
            set[datetime.date]: Set of dates (year, month, day)
        """
        return {rec.start_date.date for rec in models}

    def _map_record_keys_to_flags(self) -> dict[str, set]:
        """
        Map record keys (e.g. `unit`, `value`, `creationDate`) for each flag.

        For example:

        ```python
        HKCategoryTypeIdentifierAppleStandHour         {'startDate', 'device', 'sourceName', 'endDate', ...}
        HKCategoryTypeIdentifierAudioExposureEvent     {'startDate', 'device', 'sourceName', 'endDate', ...}
        HKQuantityTypeIdentifierActiveEnergyBurned     {'startDate', 'device', 'sourceName', 'endDate', ...}
        ```

        Returns:
            dict[str, set]: Dictionary with flags as keys and set of record keys as values
        """
        return {
            flag: {
                k
                for rec in self.records["HKQuantityTypeIdentifierActiveEnergyBurned"]
                for k in rec.attrib.keys()
            }
            for flag in self.flags
        }

    @overload
    def get_flag_records(self, flag: str) -> ParsedData: ...
    @overload
    def get_flag_records(self, flag: list[str]) -> dict[str, ParsedData]: ...
    @timeit
    def get_flag_records(
        self, flag: str | list[str]
    ) -> ParsedData | dict[str, ParsedData]:
        """
        Get parsed data based on the given flag.

        Args:
            flag (str | list[str]): Flag to parse the records (e.g., `"HKQuantityTypeIdentifierHeartRate"`)

        Returns:
            ParsedData | dict[str, ParsedData]: Parsed data based on the flag(s)
        """

        def _get_parsed_data(flag: str) -> ParsedData:
            sources = self.get_sources(flag=flag)
            devices = self.get_devices(flag=flag)
            models = self._build_models(flag=flag)
            dates = self._get_dates(models=models)

            records = pd.DataFrame([model.model_dump() for model in models])
            return ParsedData(
                flag=flag,
                sources=sources,
                devices=devices,
                dates=dates,
                records=records,
            )

        if isinstance(flag, str):
            return _get_parsed_data(flag=flag)

        elif isinstance(flag, list):
            return {f: _get_parsed_data(flag=f) for f in flag}

    def get_devices(self, flag: str | None = None) -> list[str] | dict[str, list[str]]:
        """
        Get devices for each flag or for a given flag.

        Args:
            flag (str, optional): Get devices for the given flag, defaults to None

        Returns:
            list[str] | dict[str, list[str]]: Dictionary with flags as keys and list of devices as values
        """

        def _get_device_name(rec: ET.Element) -> str:
            """
            Get the device name from the record.

            For example:
            `'<<HKDevice: 0x9999zz000>, name:Apple Watch, manufacturer:Apple Inc., model:Watch, hardware:Watch6,7, software:10.6>'`

            Should return:
            `'Apple Watch (Watch6,7; 10.6)'`

            Args:
                rec (ET.Element): Record from the export.xml file

            Returns:
                str: Device name with model and software version
            """
            device_info = rec.attrib["device"].split(", ")
            name = device_info[1].split(":")[1]
            model = device_info[4].split(":")[1]
            try:
                software = device_info[5].split(":")[1].strip(">")
                return f"{name} ({model}; {software})"
            except IndexError:
                logger.debug(f"No software version found for {name} ({model})")
                return f"{name} ({model})"

        if flag:
            return sorted(
                {
                    _get_device_name(rec)
                    for rec in self.records[flag]
                    if "device" in rec.attrib
                }
            )
        else:
            return {
                flag: sorted(
                    {
                        _get_device_name(rec)
                        for rec in self.records[flag]
                        if "device" in rec.attrib
                    }
                )
                for flag in self.flags
            }

    def get_sources(self, flag: str | None = None) -> list[str] | dict[str, list[str]]:
        """
        Get sources for each flag or for a given flag.

        Args:
            flag (str, optional): Get sources for the given flag, defaults to None

        Returns:
            list[str] | dict[str, list[str]]: Dictionary with flags as keys and list of sources as values
        """
        if flag:
            return sorted({rec.attrib["sourceName"] for rec in self.records[flag]})
        else:
            return {
                flag: sorted({rec.attrib["sourceName"] for rec in self.records[flag]})
                for flag in self.flags
            }

    @staticmethod
    def write_csv(data: ParsedData, filename: str) -> None:
        """
        Write the parsed data to a CSV file.

        Args:
            data (ParsedData): Parsed data to write to CSV file
            filename (str): Filename to write the data

        Raises:
            RecordsMissing: No records to write to CSV file
            FileTypeInvalid: File type is incorrect
        """
        if data.records is None:
            raise MissingRecords

        filepath = Path(filename)
        if filepath.suffix != ".csv":
            raise InvalidFileFormat(filepath.suffix)

        data.records.to_csv(filepath, index=False)
        logger.info(f"Parsed data written to {filepath}")

    @timeit
    def export(self, dir_name: str) -> None:
        """
        Export all parsed data to multiple CSV files.

        Args:
            dir_name (str): Directory name to export the CSV files to, defaults to current directory
        """
        export_dir = Path(dir_name)
        export_dir.mkdir(exist_ok=True)

        logger.info(f"Exporting parsed data to {export_dir}...")

        for n, flag in enumerate(self.flags):
            try:
                parsed = self.get_flag_records(flag=flag)
            except Exception:
                logger.error(f"Error parsing {flag=}")
                continue

            filename = f"{dir_name}/{flag}.csv"
            self.write_csv(data=parsed, filename=filename)

            logger.info(f"Exported {n + 1}/{len(self.flags)} flags to {filename}")



from apple_health_parser.utils.parser import Parser


def test_parsed_data(parser: Parser) -> None:
    flag = "HKQuantityTypeIdentifierHeartRate"
    parsed = parser.get_flag_records(flag)

    assert parsed.flag == flag
    assert len(parsed.dates) == 2
    assert len(parsed.sources) == 1
    assert len(parsed.records) == 2
